TRAINABLE_MODEL_TYPES = [
    'visual-classifier', 'visual-detector', 'visual-segmenter', 'visual-embedder', 'clusterer',
    'text-classifier', 'embedding-classifier', 'text-to-text'
]
MAX_MODEL_PREDICT_INPUTS = 128
MODEL_EXPORT_TIMEOUT = 1800
MIN_RANGE_SIZE = 4194304  # 4MB
MAX_RANGE_SIZE = 314572800  # 300MB
MIN_CHUNK_SIZE = 131072  # 128KB
MAX_CHUNK_SIZE = 10485760  # 10MB
RANGE_SIZE = 31457280  # 30MB
CHUNK_SIZE = 1048576  # 1MB
